#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark.ml.common import inherit_doc
from pyspark.ml.util import MLReadable, JavaMLReader, _jvm


class PipelineJavaMLReader(JavaMLReader):
    """
    MLPipeline JavaML Reader
    """

    def __init__(self, owner_clazz, class_name):
        self._className = class_name
        super(PipelineJavaMLReader, self).__init__(owner_clazz)

    def _java_loader_class(self, clazz):
        return self._className

    def _load_java_obj(self, clazz):
        """Load the peer Java object of the ML instance."""
        java_class = self._java_loader_class(clazz)
        java_obj = _jvm()
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)
        return java_obj





